#=======================================================================
#	$Id: WebEdit.pm,v 1.8 2007/09/04 10:34:06 pythontech Exp $
#	CGI application to show / edit resources in triple database.
#	Copyright (C) 2005-2007  Python Technology Limited
#
#	This program is free software; you can redistribute it and/or
#	modify it under the terms of the GNU General Public License
#	as published by the Free Software Foundation; either version 2
#	of the License, or (at your option) any later version.
#
#	This program is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with this program; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  
#	02111-1307, USA.
#=======================================================================
package PythonTech::Triples::WebEdit;
use strict;
use PythonTech::Conv qw(hq uq);
use Digest::MD5 qw(md5_hex);

my $nl = "\n";

#-----------------------------------------------------------------------
#	Create a new triples-editor
#	  db		Triple store
#	  dbroot	Root of triple storage file system tree
#	  editing_users	List of user ids allowed to edit
#	  viewing_users	List of user ids allowed to view ('*' = all)
#	  qmap		Qprefix map for user interface
#	  dbqmap	Qprefix map for underlying triplestore
#	  baseUrl	Base to make URL from relative URIs
#	  templater	(optional) template library
#-----------------------------------------------------------------------
sub new {
    my $class = shift;
    my $self = {@_};
    if (! $self->{'db'}) {
	if (! $self->{'dbroot'}) {
	    die "$class: Neither db nor dbroot configured\n";
	}
    }
    bless $self, $class;
    return $self;
}

#-----------------------------------------------------------------------
#	Handle a web query
#-----------------------------------------------------------------------
sub webquery {
    my($self, $query) = @_;
    my $qmap = $self->{'qmap'};

    #--- Access control
    # If open_browsing anyone can view, without login
    # Otherwise need login to identify themselves; they may then have
    # permission to view and maybe edit, depending on permission
    # lists.
    my $user;
    my $canedit;
    my $method = $query->request_method;
    if ($method eq 'GET') {
	# Browsing only
	if ($self->{'open_browsing'} ||	# Legacy
	    $self->{'viewing_users'} eq '*') {
	    # Anyone can view.  Is there a possibility of editing?
	    my $user = $query->session->get('userName');
	    if (defined $user) {
		if (&user_in_list($user, $self->{'editing_users'})) {
		    $canedit = 1;
		}
	    } else {
		# Not logged in
	    }
	} else {
	    $user = $query->require_login
		or return;
	    die "Viewing not authorized\n"
		unless &user_in_list($user, $self->{'viewing_users'});
	    if (&user_in_list($user, $self->{'editing_users'})) {
		$canedit = 1;
	    }
	}
    } else {
	# Trying to POST (edit) - must login
	my $user = $query->require_login 
	    or return;
	die "Editing not authorized\n"
	    unless &user_in_list($user, $self->{'editing_users'});
	$canedit = 1;
    }

    #--- Multi-database access
    # Currently hardcoded to DBHash
    if ($self->{'dbroot'}) {
	my $db = $query->param('db'); # relative path of database file
	die "Triple db query parameter not defined\n"
	    unless defined $db;
	die "Detected illegal .. in relative path [$db]\n"
	    unless (index($db,"../") < 0); 
	my $dbpath = $self->{'dbroot'} . $db;
	eval "require PythonTech::Triples::DBHash";
	die $@ if $@;
	my $hashdb = PythonTech::Triples::DBHash->new(
	    filename => "$dbpath",
	    access => 'w',
	);
	$self->{'db'} = $hashdb;
    }
    my $xuri = $query->param('u');
    my $uri = $qmap ? $qmap->decode($xuri) : $xuri;
    my $backurl = $query->param('back');

    if ($method eq 'GET') {
	if (defined $query->param('search')) {
	    $self->search($query);
	} else {
	    $self->show($query, $uri,$canedit);
	}
    } elsif ($method eq 'POST') {
	if ($query->param('delete')) {
	    $self->delete($query, $uri);
	    $self->show($query, $uri,$canedit);
	} elsif ($query->param('add')) {
	    $self->add($query, $uri,0);
	    $self->show($query, $uri,$canedit);
	} elsif ($query->param('add.lit')) {
	    $self->add($query, $uri,1);
	    $self->show($query, $uri,$canedit);
	} else {
	    die "Unknown POST request\n";
	}
    } else {
	die "Unsupported HTTP request method $method\n";
    }
}

#-----------------------------------------------------------------------
#	Show details of a URI
#-----------------------------------------------------------------------
sub show {
    my($self, $query,$uri,$editable) = @_;
    my $db = $self->{'db'};
    my $qmap = $self->{'qmap'};
    my $dbqmap = $self->{'dbqmap'};
    my $resp = $query->response;
    my $selfurl = $query->url(-path_info=>1);
    my $backurl = $query->param('back');

    # Template data
    my $data = {
	'selfurl' => $selfurl,
	'editable' => $editable,
	'backurl' => $backurl,
	&_extra_vars($query),
    };

    # External (qmapped) representation of URI
    my $xuri = $qmap ? $qmap->encode($uri) : $uri;
    $data->{'xuri'} = $xuri;

    my $url;			# URI may correspond to a URL

    if ($xuri ne $uri) {
	# URI was abbreviated to a qname
	my $qprefix = $xuri =~ m!^([-\w]+):!;
	my $qexp = $qmap->decode("$qprefix:");
#	$resp->write("<h4>$qprefix: = ".hq($qexp)."</h4>\n");
    } elsif ($dbqmap) {
	$url = $dbqmap->decode($uri);
	$uri = $dbqmap->encode($uri);
	my $last = $dbqmap->lastq;
	if (defined $last) {
#	    $resp->write("<h4> ".hq($last)." = ".hq($dbqmap->decode($last))."</h4>\n");
	}
    } elsif ($uri =~ /^http:/) {
	$url = $uri;
    } elsif ($uri !~ /^[-\w]+:/ && $self->{'baseUrl'}) {
	$url = $self->{'baseUrl'} . $uri;
    }
    if (defined $url) {
	$data->{'url'} = $url;
    }

    my $backurl = $query->param('back');

    my @tpred;
    # Forward relations
    # N.B. s2p may be slow for large triplestores without index_s2p
    # 'iterate' should do the best and avoid multiple queries.
    my %by_pred;
    $db->iterate($uri,undef,undef, sub {
	my($subj,$pred,$obj) = @_;
	$by_pred{$pred}->{$obj} = 1;
	undef;			# Continue
    });
    foreach my $p (keys %by_pred) {
	my $xp = $qmap ? $qmap->encode($p) : $p;
	my @o = sort keys %{$by_pred{$p}};
	my @tobj;
	foreach my $o (@o) {
	    if (my($str) = $o =~ /^=(.*)/s) {
		my $sel = "=" . md5_hex($str);
		push @tobj, {'is_literal' => 1, 'literal' => $str, 'sel' => $sel};
	    } else {
		my $xo = $qmap ? $qmap->encode($o) : $o;
		my $sel = $xo;
		push @tobj, {'xuri' => $xo, 'sel' => $sel}; 
	    }
	}
	@tobj = sort _valcmp @tobj;
	$tobj[0]->{'is_first'} = 1;
	push @tpred, {'is_reverse' => 0, 'pred' => $xp, 'values' => \@tobj};
    }
    

    # Reverse relations
    %by_pred = ();
    $db->iterate(undef,undef,$uri, sub {
	my($subj,$pred,$obj) = @_;
	$by_pred{$pred}->{$subj} = 1;
	undef;
    });
    foreach my $p (keys %by_pred) {
	my $xp = $qmap ? $qmap->encode($p) : $p;
	my @s = sort keys %{$by_pred{$p}};
	my @tsubj;
	foreach my $s (@s) {
	    if (my($str) = $s =~ /^=(.*)/s) {
		# Should never happen
		push @tsubj, {'is_literal' => 1, 'literal' => $str};
	    } else {
		my $xs = $qmap ? $qmap->encode($s) : $s;
		push @tsubj, {'xuri' => $xs, 'sel' => $xs};
	    }
	}
	@tsubj = sort _valcmp @tsubj;
	$tsubj[0]->{'is_first'} = 1;
	push @tpred, {'is_reverse' => 1, 'pred' => $xp, 'values' => \@tsubj};
    }
    @tpred = sort _relcmp @tpred;
    $data->{'relations'} = \@tpred;

    # Format data as HTML
    if (my $tmpl = $self->{'templater'}) {
	my $html = $tmpl->process_file('PythonTech-WebEdit-show.ptt',
				       $data);
	$resp->write($html);
    } else {
	my $hmain = $self->_format($data);

	$resp->write($query->start_html(-title=>$xuri));
	$resp->write($hmain);
	$resp->write($query->end_html);
    }
}

sub _valcmp {
    $a->{'is_literal'} <=> $b->{'is_literal'} ||
	$a->{'is_literal'} ?
	    ($a->{'literal'} cmp $b->{'literal'}) :
		($a->{'xuri'} cmp $b->{'xuri'});
}

sub _relcmp {
    $a->{'is_reverse'} <=> $b->{'is_reverse'} ||
	($a->{'pred'} ne 'rdf:type') <=> ($b->{'pred'} ne 'rdf:type') ||
	    $a->{'pred'} cmp $b->{'pred'};
}

#-----------------------------------------------------------------------
#	Search for a string in subjects and literals
#-----------------------------------------------------------------------
sub search {
    my($self, $query) = @_;
    my $search = $query->param('search');
    if ($search !~ /\S/) {
	die "Empty search string\n";
    }
    my $resp = $query->response;
    my $selfurl = $query->url(-path_info=>1);
    my $backurl = $query->param('back');
    my $data = {
	'selfurl' => $selfurl,
	'search' => $search,
	'backurl' => $backurl,
	&_extra_vars($query),
	'subjects' => [],
	'literals' => [],
    };

    # Look through everything
    my %matching_subjects;
    my %subjects_with_matching_literals;
    $self->{'db'}->iterate(undef,undef,undef, sub {
	my($subj,$pred,$obj) = @_;
	if (index($subj,$search) >= 0) {
	    $matching_subjects{$subj} = 1;
	}
	if ($obj =~ /^=.*\Q$search\E/) {
	    $subjects_with_matching_literals{$subj} = 1;
	}
	return;
    });
    foreach my $subj (sort keys %matching_subjects) {
	my $url = "$selfurl?u=" . uq($subj) . $data->{'queryextra'};
	push @{$data->{'subjects'}}, {
	    url => $url, 
	    xuri => $subj,
	};
    }
    foreach my $subj (sort keys %subjects_with_matching_literals) {
	my $url = "$selfurl?u=" . uq($subj) . $data->{'queryextra'};
	push @{$data->{'literals'}}, {
	    url => $url, 
	    xuri => $subj,
	};
    }

    if (my $tmpl = $self->{'templater'}) {
	my $html = $tmpl->process_file('PythonTech-WebEdit-matches.ptt',
				       $data);
	$resp->write($html);
    } else {
	$resp->write($query->start_html(-title=>"Search for [$search]"));
	my $html = $self->_format_matches($data);
	$resp->write($html);
	$resp->write($query->end_html);
    }
}

#-----------------------------------------------------------------------
#	Format as HTML page, without using templates
#	  selfurl	URL of this script
#	  editable	True if user may edit data
#	  xuri		External representation of URI
#	  url		(opt) URL for viewing resource
#	  back		(opt) URL for returning
#	  queryextra	Extra terms for URL query (db, back)
#	  HTML:formextra	Hidden fields (db, back) for form
#	  relations	Array of hashes for each relation:
#			  is_reverse	True for (subj,pred,THIS)
#			  pred		Predicate
#			  values	Array of hashes
#			    is_first	  True for first value
#			    sel		  Selector
#			    is_literal	  True for literal object
#			    literal	  Literal object value
#			    xuri	  External subject/object URI
#-----------------------------------------------------------------------
sub _format {
    my($self, $data) = @_;
    my $selfurl = $data->{'selfurl'};
    my $editable = $data->{'editable'};
    my $huri = hq($data->{'xuri'});
    my $queryextra = $data->{'queryextra'};
    my $hformextra = $data->{'HTML:formextra'};
    my $html;

    if ($editable) {
	$html .= '<form method="POST">'.$nl.
	    '<input type="hidden" name="u" value="'.$huri.'" />'.$nl.
		$hformextra;
    }
    $html .= '<table border="1">'.$nl;
    # Title: URI
    $html .= '<tr><th colspan="2">';
    if ($data->{'url'}) {
	$html .= '<a href="'.hq($data->{'url'}).'">'.$huri.'</a>';
    } else {
	$html .= $huri;
    }
    $html .= '</th></tr>'.$nl;
    # Each predicate, by relation (forward then reversed)
    foreach my $rel (@{$data->{'relations'}}) {
	my @vals = @{$rel->{'values'}};
	my $rev = $rel->{'is_reverse'};
	foreach my $val (@vals) {
	    $html .= '<tr>';
	    if ($val->{'is_first'}) {
		$html .= '<td rowspan="'.scalar(@vals).'">';
		if ($rev) {
		    $html .= '~ ';
		}
		$html .= hq($rel->{'pred'});
		$html .= '</td>';
	    }
	    $html .= '<td>';
	    if ($editable) {
		my($name,$value);
		if ($rev) {
		    $name = 'sp';
		    $value = "$val->{'sel'} $rel->{'pred'}";
		} else {
		    $name = 'po';
		    $value = "$rel->{'pred'} $val->{'sel'}";
		}
		$html .= '<input type="checkbox" name="'.$name.
		    '" value="'.hq($value).'" />';
	    }
	    if ($val->{'is_literal'}) {
		my $str = $val->{'literal'};
		$html .= '<span class="literal">'.hq($str).'</span>';
	    } else {
		my $xuri = $val->{'xuri'};
		my $url = "$selfurl?u=".uq($xuri).$queryextra;
		$html .= '<a href="'.hq($url).'">'.hq($xuri).'</a>';
	    }
	    $html .= '</td>';
	    $html .= '</tr>'.$nl;
	}
    }
    $html .= '</table>'.$nl;
    if ($editable) {
	$html .= '<input type="submit" name="delete" value="Delete selected" />
<dl>
<dt>Predicate</dt>
  <dd><input type="text" name="p" width="60" /></dd>
<dt>Object</dt>
  <dd><input type="text" name="o" width="60" />
   <input type="submit" name="add" value="Add" />
  </dd>
  <dd><textarea name="ol" cols="60" rows="4"></textarea>
   <input type="submit" name="add.lit" value="Add (literal)" />
  </dd>
</dl>
</form>';
    }
    if (1) {
	$html .= $self->_format_search($data);
    }
    return $html;
}

#-----------------------------------------------------------------------
#	Format search results, without using templates
#	  selfurl	URL of this script
#	  search	String searched for
#	  back		(opt) URL for returning
#	  queryextra	Extra terms for URL query (db, back)
#	  HTML:formextra	Hidden fields (db, back) for form
#	  subjects	Array of hashes for matching subjects
#			  url		URL for viewing resource
#			  xuri		External resource URI
#	  literals	Array of hashes for subjects with matching literals
#			  url		URL for viewing resource
#			  xuri		External resource URI
#-----------------------------------------------------------------------
sub _format_matches {
    my($self, $data) = @_;
    my $selfurl = $data->{'selfurl'};
    my $hsearch = hq($data->{'search'});
    my $html;

    $html .= '<table border="1">'.$nl;
    if ($data->{'backurl'}) {
	my $hback = hq($data->{'backurl'});
	$html .= '<tr><th colspan="2">Model context: '.
	    '<a href="'.$hback.'">'.$hback.'</a></th></tr>'.$nl;
    }

    if (@{$data->{'subjects'}} == 0) {
	$html .= '<tr><th colspan="2">'.
	    'No subjects matching search for ['.$hsearch.']'.
		'</th></tr>'.$nl;
    } else {
	$html .= '<tr><th>'.
	    'Subjects matching search<br />for ['.$hsearch.']:'.
		'</th>'.
		    '<td><ul>'.$nl;
	foreach my $rec (@{$data->{'subjects'}}) {
	    $html .= '<li><a href="'.hq($rec->{'url'}).'">'.
		hq($rec->{'xuri'}).
		    '</a></li>'.$nl;
	}
	$html .= '</ul></td></tr>'.$nl;
    }

    if (@{$data->{'literals'}} == 0) {
	$html .= '<tr><th colspan="2">'.
	    'No literals matching search for ['.$hsearch.']'.
		'</th></tr>'.$nl;
    } else {
	$html .= '<tr><th>'.
	    'Subjects with literals<br />matching search for ['.$hsearch.']:'.
		'</th>'.
		    '<td><ul>'.$nl;
	foreach my $rec (@{$data->{'literals'}}) {
	    $html .= '<li><a href="'.hq($rec->{'url'}).'">'.
		hq($rec->{'xuri'}).
		    '</a></li>'.$nl;
	}
	$html .= '</ul></td></tr>'.$nl;
    }

    $html .= '</table>'.$nl;
    if (1) {
	$html .= $self->_format_search($data);
    }
    return $html;
}

#-----------------------------------------------------------------------
#	Format search box without using templates
#	  selfurl		URL of script
#	  HTML:formextra	Hidden fields (db, back) for form
#-----------------------------------------------------------------------
sub _format_search {
    my($self, $data) = @_;
    my $html = ('<p>'.$nl.
		'<form method="GET">'.$nl.
		$data->{'HTML:formextra'} .
		'<input type="text" name="u" value="" />'.$nl.
		'<input type="submit" value="Go to URI" />'.$nl.
		'</form>'.$nl);
    $html .= '<br />'.$nl;
    $html .= ('<form method="GET">'.$nl.
	      $data->{'HTML:formextra'} .
	      '<input type="text" name="search" value="" />'.$nl.
	      '<input type="submit" value="Subject and literal search" />'.$nl.
	      '</form>'.$nl);
    $html .= '</p>'.$nl;
    return $html;
}

#-----------------------------------------------------------------------
#	Return 'queryextra' and 'HTML:formextra'
#-----------------------------------------------------------------------
sub _extra_vars {
    my($query) = @_;
    my $queryextra = '';
    my $hformextra = '';
    foreach my $var ('db','back') {
	my $value = $query->param($var);
	if (defined $value) {
	    $queryextra .= "&$var=" . uq($value);
	    $hformextra .= '<input type="hidden" '.
		'name="'.$var.'" '.
		    'value="'.hq($value).'" />'.$nl;
	}
    }
    return ('queryextra' => $queryextra,
	    'HTML:formextra' => $hformextra);
}

#-----------------------------------------------------------------------
#	Delete some statements
#-----------------------------------------------------------------------
sub delete {
    my($self, $query,$uri) = @_;
    my $db = $self->{'db'};
    my $qmap = $self->{'qmap'};

    # Delete forward relations
    my @po = $query->param('po');
    foreach (@po) {
	my($xp,$xo) = split / /;
	my $p = $qmap ? $qmap->decode($xp) : $xp;
	if ($xo =~ /^=/) {
	    # Literal as MD5 checksum
	    foreach my $lit ($db->sp2o($uri,$p)) {
		if ($lit =~ /^=/ &&
		    md5_hex(substr($lit,1)) eq substr($xo,1)) {
		    $db->remove($uri,$p,$lit);
		}
	    }
	} else {
	    # Resource URI
	    my $o = $qmap ? $qmap->decode($xo) : $xo;
	    $db->remove($uri,$p,$o);
	}
    }
    # Delete backward relations
    my @po = $query->param('sp');
    foreach (@po) {
	my($xs,$xp) = split / /;
	my $s = $qmap ? $qmap->decode($xs) : $xs;
	my $p = $qmap ? $qmap->decode($xp) : $xp;
	$db->remove($s,$p,$uri);
    }
    $db->commit;
}

#-----------------------------------------------------------------------
#	Add a statement
#-----------------------------------------------------------------------
sub add {
    my($self, $query,$uri,$literal) = @_;
    my $db = $self->{'db'};
    my $qmap = $self->{'qmap'};
    my $xp = $query->param('p');
    die "Malformed URI\n"
	if $xp =~ /\s/;
    my $p = $qmap ? $qmap->decode($xp) : $xp;
    my $o;
    if ($literal) {
	$o = "=" . $query->param('ol');
    } else {
	my $xo = $query->param('o');
	die "Malformed URI\n"
	    if $xo =~ /\s/;
	$o = $qmap ? $qmap->decode($xo) : $xo;
    }
    $db->add($uri,$p,$o);
    $db->commit;
}

#-----------------------------------------------------------------------
#	Utility routines
#-----------------------------------------------------------------------
sub user_in_list {
    my($user, $list) = @_;
    if (! defined $list) {
	# No list, so nobody in it
	return undef;
    } elsif ($list eq '*') {
	# Means everybody
	return 1;
    } elsif (ref($list) eq 'ARRAY') {
	# Simple list of usernames
	return (grep {$user eq $_} @$list) != 0;
    } elsif (ref($list) eq 'CODE') {
	# Some algorithm
	return $list->($user);
    } else {
	die "user_in_list: Unrecognised list: $list";
    }
}

1;
